'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var stream = _interopDefault(require('stream'));
var childProcess = _interopDefault(require('child_process'));
var util = _interopDefault(require('util'));

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(source, true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(source).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

const {
  Stream
} = stream;
class Capturer {
  constructor(o) {
    _defineProperty(this, "text", '');

    const {
      input,
      pipe
    } = Object(o);
    const m = 'Capturer';
    if (!(input instanceof Stream)) throw new Error(`${m}: input not stream: ${typeof input}`);
    this.input = input;

    if (pipe) {
      if (!(pipe instanceof Stream)) throw new Error(`${m}: pipe not stream: ${typeof pipe}`);
      this.pipe = pipe;
    }
  }

  capture() {
    var _this = this;

    return _asyncToGenerator(function* () {
      const {
        input,
        pipe
      } = _this;
      return new Promise((resolve, reject) => input.on('data', text => _this.saveText(text).catch(reject)).once('close', () => resolve(_this.text)).setEncoding('utf8') + (pipe && input.pipe(pipe)));
    })();
  }

  saveText(text) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      _this2.text += text;
    })();
  }

}

const {
  spawn
} = childProcess;
class SpawnShim {
  constructor(o) {
    const {
      args,
      options,
      killTimeout,
      name,
      debug
    } = Object(o);
    const {
      killTimeout0
    } = SpawnShim;
    this.m = String(name || 'SpawnShim');
    const isString = typeof args === 'string';
    const isArray = Array.isArray(args);
    const cmd = isString ? args : isArray && args[0];
    cmd && (this.cmd = cmd);
    isArray && (this.args = args.slice(1)) || !isString && args && (this.args = args);
    options && (this.options = options);
    this.killTimeout = killTimeout >= 0 ? +killTimeout : killTimeout0;
    debug && (this.debug = true) && this.constructor === SpawnShim && console.log(`${this.m} constructor: ${util.inspect(this, {
      colors: true,
      depth: null
    })}`);
  }

  spawn() {
    const {
      cmd,
      args,
      options
    } = this;
    const cp = this.cp = spawn(cmd, args, options); // may throw

    const promise = new Promise((resolve, reject) => {
      let e;

      const exitListener = () => cp.removeListener('exit', exitListener).removeListener('error', exitListener) + this._exitListener().catch(reject);

      cp.once('close', (status, signal) => resolve({
        e,
        status,
        signal
      })).on('error', ee => !e && (e = ee)) // subsequent errors are ignored
      .once('exit', exitListener).once('error', exitListener);
    });
    return {
      cp,
      promise
    };
  }

  _exitListener() {
    var _this = this;

    return _asyncToGenerator(function* () {
      const {
        debug
      } = _this;
      _this.hadExit = true;
      debug && console.log(`${_this.m} process exited`);
    })();
  }

  abortProcess() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      const {
        cp,
        hadExit,
        debug,
        killTimeout
      } = _this2;
      debug && console.log(`${_this2.m} abortProcess: cp: ${!!cp} hadExit: ${!!hadExit} killed: ${Object(cp).killed}`);
      if (!cp || hadExit) return;
      let cancelTimer;
      yield Promise.all([new Promise((resolve, reject) => {
        const timer = killTimeout && setTimeout(() => _this2._doSigKill(cp).then(resolve, reject), killTimeout);

        cancelTimer = () => (timer && clearTimeout(timer)) + resolve();
      }), new Promise((resolve, reject) => {
        const onProcessExit = () => cp.removeListener('exit', onProcessExit).removeListener('error', onProcessExit) + cancelTimer() + resolve();

        cp.once('exit', onProcessExit).once('error', onProcessExit);

        _this2._maybeKill(cp);
      })]);
      debug && console.log(`${_this2.m} abortProcess: exit complete: hadExit: ${_this2.hadExit}`);
    })();
  }

  _maybeKill(cp) {
    const {
      debug
    } = this;
    const {
      killed
    } = cp;

    if (!killed) {
      debug && console.log(`${this.m} invoking cp.kill…`);
      cp.kill();
    }
  }

  _doSigKill(cp) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const {
        debug
      } = _this3;
      debug && console.log(`${_this3.m} sending SIGKILL…`);
      cp.kill('SIGKILL'); // kill -9
    })();
  }

}

_defineProperty(SpawnShim, "killTimeout0", 3e3);

class SpawnPipe extends SpawnShim {
  constructor(o) {
    super(o);
    const {
      capture,
      stderrFails
    } = Object(o);
    const {
      options: options0
    } = this;
    const {
      maxBuffer0
    } = SpawnPipe;
    capture && (this.capture = true);
    stderrFails && (this.stderrFails = true);
    const options = options0 != null ? options0 : this.options = {}; // where to store stdio value

    const {
      maxBuffer,
      silent,
      stdio: stdio0
    } = options;

    if (options0 != null) {
      delete options.maxBuffer;
      delete options.silent;
    }

    this.maxBuffer = maxBuffer >= 0 ? +maxBuffer : maxBuffer0;
    silent && (this.silent = true);
    const defOut = silent ? 'pipe' : 'inherit';
    const stdio = options.stdio = Array.isArray(stdio0) ? stdio0 : typeof stdio0 === 'string' ? new Array(3).fill(stdio0) : ['ignore', defOut, defOut];

    if (capture) {
      stdio[1] = stdio[2] = 'pipe';
      if (stdio[0] === undefined) stdio[0] = 'ignore';
    } else if (stderrFails) {
      stdio[2] = 'pipe';
      if (stdio[0] === undefined) stdio[0] = 'ignore';
      if (stdio[1] === undefined) stdio[1] = 'ignore';
    }
  }

  startCapture() {
    var _this = this;

    return _asyncToGenerator(function* () {
      const {
        capture,
        stderrFails
      } = _this;
      return capture ? _this.doCapture() : stderrFails ? _this.doStderrFails() : {};
    })();
  }

  doCapture() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      const {
        cp,
        silent,
        stderrFails,
        maxBuffer
      } = _this2;
      const [stdout, stderr] = yield Promise.all(['stdout', 'stderr'].map(stream => new Capturer({
        input: cp[stream],
        pipe: !silent && process[stream],
        maxBuffer
      }).capture()));
      const isStderr = stderrFails && !!stderr;
      return {
        stdout,
        stderr,
        isStderr
      };
    })();
  }

  doStderrFails() {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const {
        cp: {
          stderr: stream
        }
      } = _this3;
      const stderr = yield new Promise((resolve, reject) => stream.once('data', text => text && resolve(text)).once('close', resolve).setEncoding('utf8'));
      const isStderr = !!stderr;
      return {
        stderr,
        isStderr
      };
    })();
  }

}

_defineProperty(SpawnPipe, "maxBuffer0", 200 * 1024);

class SpawnAsync extends SpawnPipe {
  static spawnAsync(o) {
    return _asyncToGenerator(function* () {
      return new SpawnAsync(o).startSpawn();
    })();
  }

  constructor(o) {
    super(_objectSpread2({
      name: 'SpawnAsync'
    }, o));
    const {
      cpReceiver,
      echo,
      nonZeroOk,
      signalOk,
      debug
    } = Object(o);
    const {
      options
    } = this;
    const {
      timeout
    } = options;
    echo && (this.echo = true);
    nonZeroOk && (this.nonZeroOk = true);
    signalOk && (this.signalOk = true);
    cpReceiver && (this.cpReceiver = cpReceiver);

    if (timeout != null) {
      timeout > 0 && (this.timeout = +timeout);
      delete options.timeout;
    }

    debug && this.constructor === SpawnAsync && console.log(`${this.m} constructor: ${util.inspect(this, {
      colors: true,
      depth: null
    })}`);
  }

  startSpawn() {
    var _this = this;

    return _asyncToGenerator(function* () {
      return (yield Promise.all([_this.launchProcess(), _this.setTimer()]))[0];
    })();
  }

  launchProcess() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      const {
        echo,
        cpReceiver,
        debug
      } = _this2; // launch the child process

      echo && console.log(_this2.cmdString());

      const {
        cp,
        promise
      } = _this2.spawn();

      cpReceiver && (cpReceiver.cp = cp); // await child process exit

      const [{
        e,
        status,
        signal
      }, {
        stdout,
        stderr,
        isStderr
      }] = yield Promise.all([promise, _this2.startCapture()]);
      const {
        isTimeout,
        timeout,
        clearTimer,
        nonZeroOk,
        signalOk
      } = _this2;
      debug && console.log(`${_this2.m} process exit:`, {
        e,
        status,
        signal,
        stdout: stdout && stdout.length,
        stderr: stderr && stderr.length,
        isStderr,
        isTimeout,
        ss: stdout
      }); // handle timeout

      if (isTimeout) throw _this2.setErrorProps(new Error(`Process timeout: ${(timeout / 1e3).toFixed(1)} s: ${_this2.cmdString()}`));else clearTimer && clearTimer(); // handle error from child process

      const eText = stderr && _this2.trimEnd(stderr);

      if (e) throw _this2.setErrorProps(e, eText);
      if (status && !nonZeroOk || signal && !signalOk) throw _this2.getError({
        status,
        signal
      }, eText);
      if (isStderr) throw _this2.setErrorProps(new Error(`Output on standard error: ${_this2.cmdString()}: '${eText}'`), eText);
      if (stdout === undefined) return status || signal || 0;
      const result = {
        status,
        stdout,
        stderr
      };
      signal && (result.signal = signal);
      return result;
    })();
  }

  setTimer() {
    const {
      timeout
    } = this;
    return timeout && new Promise((resolve, reject) => {
      const timer = setTimeout(() => this._onTimeout().then(resolve, reject), timeout);

      this.clearTimer = () => clearTimeout(timer) + resolve();
    });
  }

  _onTimeout() {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const {
        debug
      } = _this3;
      debug && console.log(`${_this3.m} child process timeout`);
      _this3.isTimeout = true;
      return _this3.abortProcess();
    })();
  }

  trimEnd(s) {
    return typeof s === 'string' && s.endsWith('\n') ? s.slice(0, -1) : s;
  }

  getError({
    status,
    signal
  }, stderr) {
    return this.setErrorProps(new Error(`status code: ${status}${signal ? ` signal: ${signal}` : ''} '${this.cmdString()}'`), stderr, Object.assign({
      status
    }, signal ? {
      signal
    } : null));
  }

  setErrorProps(e, stderr, o) {
    const {
      cmd,
      args
    } = this;
    return Object.assign(e, {
      cmd,
      args
    }, o, stderr ? {
      stderr
    } : null);
  }

  cmdString() {
    const {
      cmd,
      args
    } = this;
    return `${cmd || ''} ${Array.isArray(args) ? args.join('\x20') : ''}`;
  }

}

function spawnCapture(_x) {
  return _spawnCapture.apply(this, arguments);
}

function _spawnCapture() {
  _spawnCapture = _asyncToGenerator(function* (o) {
    const spawnAsync = new SpawnAsync(_objectSpread2({
      stderrFails: true
    }, o, {
      capture: true,
      options: _objectSpread2({
        timeout: 3e3,
        silent: true
      }, Object(o).options)
    }));
    const result = yield spawnAsync.startSpawn();
    const {
      stdout,
      stderr
    } = result;
    result.stdout = spawnAsync.trimEnd(stdout);
    result.stderr = spawnAsync.trimEnd(stderr);
    return result;
  });
  return _spawnCapture.apply(this, arguments);
}

/*
© 2018-present Harald Rudell <harald.rudell@gmail.com> (http://www.haraldrudell.com)
This source code is licensed under the ISC-style license found in the LICENSE file in the root directory of this source tree.
*/
const spawnAsync = SpawnAsync.spawnAsync;

exports.SpawnAsync = SpawnAsync;
exports.SpawnPipe = SpawnPipe;
exports.SpawnShim = SpawnShim;
exports.spawnAsync = spawnAsync;
exports.spawnCapture = spawnCapture;
